!
!     Copyright (c) 2015, NVIDIA CORPORATION.  All rights reserved.
!
! NVIDIA CORPORATION and its licensors retain all intellectual property
! and proprietary rights in and to this software, related documentation
! and any modifications thereto.  Any use, reproduction, disclosure or
! distribution of this software and related documentation without an express
! license agreement from NVIDIA CORPORATION is strictly prohibited.
!

!          THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT
!   WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT
!   NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
!   FITNESS FOR A PARTICULAR PURPOSE.
!

!! link.f90

module link_mod
  private
  public :: link

  type link
      private

      class(*), pointer :: value => null() ! value stored in link

      type(link), pointer :: next => null()! next link in list

      contains
      procedure :: getValue    ! return value pointer
      procedure :: printLinks  ! print linked list starting with this link
      procedure :: nextLink    ! return next pointer
      procedure :: setNextLink ! set next pointer
  end type link

  interface link
      procedure constructor ! construct/initialize a link
  end interface

contains

  function nextLink(this)
  class(link) :: this
  class(link), pointer :: nextLink
    nextLink => this%next
  end function nextLink

  subroutine setNextLink(this,next)
  class(link) :: this
  class(link), pointer :: next
     this%next => next
  end subroutine setNextLink

  function getValue(this)
  class(link) :: this
  class(*), pointer :: getValue
  getValue => this%value
  end function getValue

  subroutine printLink(this)
  class(link) :: this

  select type(v => this%value)
  type is (integer)
    print *, v
  type is (character(*)) 
   print *, v(1:1)
  type is (real)
    print *, v
  class default
    stop 'printLink: unexepected type for link'
  end select

  end subroutine printLink

  subroutine printLinks(this)
  class(link) :: this
  class(link), pointer :: curr

  call printLink(this)
  curr => this%next
  do while(associated(curr))
    call printLink(curr)
    curr => curr%next
  end do

  end subroutine

  function constructor(value, next)
    class(link),pointer :: constructor
    class(*) :: value
    class(link), pointer :: next
    allocate(constructor)
    constructor%next => next
    allocate(constructor%value, source=value)
  end function constructor

end module link_mod

!
!     Copyright (c) 2015, NVIDIA CORPORATION.  All rights reserved.
!
! NVIDIA CORPORATION and its licensors retain all intellectual property
! and proprietary rights in and to this software, related documentation
! and any modifications thereto.  Any use, reproduction, disclosure or
! distribution of this software and related documentation without an express
! license agreement from NVIDIA CORPORATION is strictly prohibited.
!

!          THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT
!   WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT
!   NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
!   FITNESS FOR A PARTICULAR PURPOSE.
!

!! abstract_list.f90

module abstract_list_mod
  use link_mod
  private
  public :: list
!Isso pode ser feito por meio de um tipo abstrato. 
!Um tipo abstrato  um tipo derivado que nao pode ser instanciado. 
!Em vez disso, ele  estendido e mais definido por outro tipo. 
!A extensao do tipo tambm pode ser declarada como abstrata, mas, 
!em ltima anlise, deve ser estendida por um tipo no abstrato se 
!algum dia for ser instanciada em um programa. 
!Abaixo est um exemplo de um tipo de lista declarado como abstrato:
  type, abstract :: list
     private
     class(link),pointer :: firstLink => null() ! first link in list
     class(link),pointer :: lastLink => null()  ! last link in list
     class(link),pointer :: currLink => null()  ! list iterator
   contains
     procedure, non_overridable :: addValue     ! add class(*) to list
     procedure, non_overridable :: firstValue   ! value of first link in list
     procedure, non_overridable :: reset        ! reset list iterator
     procedure, non_overridable :: next         ! increment list iterator
     procedure, non_overridable :: currentValue ! get value from currLink
     procedure, non_overridable :: moreValues   ! more values for iterator?
     ! generic :: add => addValue, why: https://software.intel.com/en-us/forums/intel-fortran-compiler/topic/607462
     procedure(printValues), deferred :: printList ! prints values in list
  end type list

  abstract interface
    subroutine printValues(this)
      import list
      class(list) :: this
    end subroutine
  end interface

contains

  subroutine addValue(this, value)
    class(list)          :: this
    class(*)             :: value
    class(link), pointer :: newLink

    if (.not. associated(this%firstLink)) then
       this%firstLink => link(value, this%firstLink)
       this%lastLink => this%firstLink
    else
       newLink => link(value, this%lastLink%nextLink())
       call this%lastLink%setNextLink(newLink)
       this%lastLink => newLink
    end if

  end subroutine addValue

  function firstValue(this)
    class(list) :: this
    class(*), pointer :: firstValue

    firstValue => this%firstLink%getValue()

  end function firstValue

  function currentValue(this)
   class(list) :: this
   class(*), pointer :: CurrentValue
   CurrentValue => this%currLink%getValue()
  end function CurrentValue

  subroutine next(this)
   class(list) :: this
   this%currLink => this%currLink%nextLink()
  end subroutine next

  function moreValues(this)
   class(list) :: this
   logical moreValues
   moreValues = associated(this%currLink)
  end function moreValues

  subroutine reset(this)
  class(list) :: this
    this%currLink => this%firstLink
  end subroutine reset

end module abstract_list_mod

!
!----------------------------------------------------------------------------------
!

! https://software.intel.com/en-us/forums/intel-fortran-compiler/topic/607462
!
!     Copyright (c) 2015, NVIDIA CORPORATION.  All rights reserved.
!
! NVIDIA CORPORATION and its licensors retain all intellectual property
! and proprietary rights in and to this software, related documentation
! and any modifications thereto.  Any use, reproduction, disclosure or
! distribution of this software and related documentation without an express
! license agreement from NVIDIA CORPORATION is strictly prohibited.
!

!          THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT
!   WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT
!   NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR
!   FITNESS FOR A PARTICULAR PURPOSE.
!

!! integerList.f90

module integer_list_mod
  use abstract_list_mod
  private
  public :: integerList
  type, extends(list) :: integerList
   contains
     procedure :: addInteger                    ! add integer in list
     procedure :: current => currentInteger     ! get integer pointed by iterator
     procedure :: printList => printIntegerList ! print the integer list
     generic   :: add => addInteger
  end type integerList

contains

  subroutine printIntegerList(this)
    class(integerList) :: this
    class(*), pointer  :: curr

    call this%reset()
    do while(this%moreValues())
       curr => this%CurrentValue()
       select type(curr)
       type is (integer)
          print *, curr
       end select
       call this%next()
    end do
    call this%reset()
  end subroutine printIntegerList

  subroutine addInteger(this, value)
    class(integerList)    :: this
    integer               :: value
    class(*), allocatable :: v

    allocate(v,source=value)
    
    call this%addValue(v)
    
    ! Previously: this%list%add(v), but it won't work, see:
    ! https://software.intel.com/en-us/forums/intel-fortran-compiler/topic/505472
    ! https://software.intel.com/en-us/forums/intel-fortran-compiler/topic/607462
    
  end subroutine addInteger

  integer function currentInteger(this)
    class(integerList) :: this
    class(*), pointer :: v

    v => this%currentValue()
    select type(v)
    type is (integer)
      currentInteger = v
    end select
  end function currentInteger

end module integer_list_mod
!
!----------------------------------------------------------------------------------
!
program main
  use integer_list_mod
  implicit none
  integer i
  type(integerList) :: my_list
  integer values(10)
  
  do i=1, 10
     call my_list%add(i)
  enddo
  
  call my_list%printList()
  print *
  
  call my_list%reset()
  i = 1
  do while(my_list%moreValues())
     values(i) = my_list%current()
     call my_list%next()
     i = i + 1
  end do
  
  print *, values
end program main
