!An impure elemental procedure can have side effects and it can contain
! the following:
!
!Any external I/O statement (including a READ or WRITE statement 
!whose I/O unit is an external file unit number or *)
!
!A PAUSE statement
!
!A STOP statement or an ERROR STOP statement
!
!An image control statement
!
!An impure elemental procedure cannot be referenced in a context 
!that requires a procedure to be pure; for example:
!
!It cannot be called directly in a FORALL statement or
! be used in the mask expression of a FORALL statement.
!
!It cannot be called from a pure procedure. 
!Pure procedures can only call other pure procedures, 
!including one referenced by means of a defined operator, 
!defined assignment, or finalization.
!
!It cannot be passed as an actual argument to a pure procedure.
!
module my_rand_mod

integer, save :: my_rand_seed (8)


  contains
    impure elemental subroutine my_rand (r)

    ! my_rand updates module variable my_rand_seed (an array)
    ! and returns the next value of a ?pseudo? random sequence in
    ! output dummy argument r

      real, intent(out) :: r

      ! code goes here
      
    end subroutine my_rand
end module my_rand_mod

!A pure procedure must not contain the following:
!
!Any external I/O statement (including a READ or WRITE
! statement whose I/O unit is an external file unit number or *)
!
!A PAUSE statement
!
!A STOP statement
!
!An image control statement
!
!An OpenMP* directive
!
!A pure procedure has no side effects. 
!It has no effect on the state of the program, except for the following:
!
!For functions: It returns a value.
!
!For subroutines: It modifies INTENT(OUT) and INTENT(INOUT) parameters.
!
!A local variable declared in a pure procedure 
!(including variables declared in any internal procedure) must not:
!
!Specify the SAVE attribute
!
!Be initialized in a type declaration statement or a DATA statement
!
!The following variables have restricted use in pure procedures 
!(and any internal procedures):
!
!Global variables
!
!Dummy arguments with INTENT(IN) (or no declared intent)
!
!Objects that are storage associated with any part of a global variable
!
!!----------------------------------------------------------------------------
!!                              -----module------
!!----------------------------------------------------------------------------
module class_Circle
  implicit none

  private

  real :: pi = 3.1415926535897931d0 ! Class-wide private constant

  type, public :: Circle
   real :: radius

   contains
   
     procedure, private :: circle_area
     procedure, private :: circle_area2
     generic   :: area => circle_area,circle_area2


     procedure :: print => circle_print
  end type Circle

!
!Note que alteramos os parmetros fictcios de type(Circle) para class(Circle).
! Podemos ento chamar as funes da classe de forma orientada a objetos 
! como a = c%area e chamar c%print, onde o tipo  passado 
! para o primeiro argumento das funes circle_area e circle_print 
! automaticamente. 
!
  interface area
    pure module function circle_area(self) result(area)
      class(Circle), intent(in) :: self
      real :: area
    end function circle_area

    pure module function circle_area2(self,log) result(area2)
      class(Circle), intent(in) :: self
      CHARACTER(LEN=*) , intent(in) :: log
      real :: area2
    end function circle_area2
  end interface area


  interface print

    impure module subroutine circle_print(self,charx)
      class(Circle), intent(in) :: self
      CHARACTER(LEN=*) , intent(in) :: charx
      real :: area
    end subroutine circle_print

  end interface print

end module class_Circle
!!----------------------------------------------------------------------------
!!                              -----submodule------
!!----------------------------------------------------------------------------
submodule (class_Circle) class_Circle_a
implicit none
real, parameter :: pi2=2.0*4.0*atan(1.0)
real :: test

contains

!
! Note que alteramos os parmetros fictcios de type(Circle) para class(Circle).
! Podemos ento chamar as funes da classe de forma orientada a objetos 
! como a = c%area e chamar c%print, onde o tipo  passado 
! para o primeiro argumento das funes circle_area e circle_print 
! automaticamente. 
!
  pure module function circle_area(self) result(area)
    class(Circle), intent(in) :: self
    real :: area 
    !PRINT statement at (1) not allowed within PURE procedure
    !print*,'2*pi: ',pi2
    area = pi * self%radius**2
  end function circle_area

  pure module function circle_area2(self,log) result(area2)
    class(Circle), intent(in) :: self
    CHARACTER(LEN=*) , intent(in) :: log
    real :: area2
    !PRINT statement at (1) not allowed within PURE procedure
    !print*,'2*pi: ',pi2
    area2 = (pi2/2.0) * self%radius**2
  end function circle_area2
  
  impure module subroutine circle_print(self,charx)
    class(Circle), intent(in) :: self
    CHARACTER(LEN=*) , intent(in) :: charx
    real :: area
    area = self%area()  ! Call the type-bound function
    print*,'charx: ',charx
    print *, 'Circle: r = ', self%radius, ' area = ', area
    test=1
  end subroutine circle_print
  
end submodule class_Circle_a

program circle_test
  use class_Circle
  implicit none
  type(Circle) :: c     ! Declare a variable of type Circle.

  c = Circle(1.5)           ! Use the implicit constructor,initialized the radius = 1.5.
  call c%print('teste-1')  ! Call the type-bound subroutine
  c = Circle(2.0)           ! Use the implicit constructor,initialized the radius = 1.5.
  call c%print('teste-2')  ! Call the type-bound subroutine

end program circle_test
