module p2fi
  ! This module fits a quadratic function using a small neural network using
  ! stochastic gradient descent, batch gradient descent, and minibatch gradient
  ! descent.
  use nf, only: dense,flatten, input, network,activation_function
  use nf_dense_layer, only: dense_layer
  use nf_flatten_layer, only: flatten_layer
  use nf_optimizers, only: sgd, rmsprop, adam, adagrad  
  use nf_activation, only: elu, celu,linear,leaky_relu,gaussian,sigmoid,tanhf,relu,exponential
  use Class_Statistic, only:AVE,SD

  implicit none
  private
  type(network) :: net(11)

  ! Training parameters
  integer, parameter :: num_epochs    = 100
  
  integer, parameter :: train_size    = 7000
  integer, parameter :: test_size     = 1000
  
  integer, parameter :: batch_size    = 100
  real   , parameter :: learning_rate = 0.001
  real   , parameter :: decay_rate    = 0.9
  real   , parameter :: beta1         = 0.85
  real   , parameter :: beta2         = 0.95
  real   , parameter :: epsilon       = 1e-8
  character(Len=528) :: path_in,path_out
  character(Len=528) :: filename_train,filename_test

  ! Input and output data
  real, allocatable :: xr(:),x(:), y(:)       ! training data
  real, allocatable :: xtest(:), ytest(:) ! testing data

  integer :: i, n
  real    :: ave_x_train, ave_y_train, sd_x_train, sd_y_train 
  public :: init,Run,Finalize

contains 
 subroutine init(val)
  integer, parameter :: sp = selected_real_kind(6,37) ! single precision
  real(sp), intent(in) :: val

  print '("Fitting quadratic function")'
  print '(60("="))'
  ! xtest and ytest as 1-D arrays
  allocate(xtest(test_size), ytest(test_size))
  ! x and y as 1-D arrays
  allocate(x(train_size),xr(train_size), y(train_size))

  path_in='/cygdrive/d/paulo.kubota/pgmet/isabella/neural-fortran/example/isabella/datain'
  path_out='/cygdrive/d/paulo.kubota/pgmet/isabella/neural-fortran/example/isabella/datain'
  filename_train='datatrain.dat'
  filename_test='datatest.dat'

  OPEN(10,file=TRIM(path_in)//'/'//trim(filename_train),status='old',action='read',form='formatted') 
  read(10,'(1X)')
  do i=1, train_size
    read (10,*)x(i),y(i)
  end do

  OPEN(20,file=TRIM(path_in)//'/'//trim(filename_test),status='old',action='read',form='formatted') 
  read(20,'(1X)')
  do i=1, test_size
    read (20,*)xtest(i),ytest(i)
  end do

  print '("Simple")'
  print '(60("="))'  

 end subroutine init

 subroutine Run(val)
  integer, parameter :: sp = selected_real_kind(6,37) ! single precision
  real(sp), intent(in) :: val

  
  ! Generating the dataset
  ave_y_train=ave(y)
  sd_y_train =sd(y)
  ave_x_train=ave(x)
  sd_x_train =sd(x)

  do i = 1, train_size
    x(i) = (x(i)-ave_x_train)/(2*sd_x_train)!ABS(x(i)/MAXVAL(ABS(x)))
    y(i) = (y(i)-ave_y_train)/(2*sd_y_train)
  end do

  !xtest = [(real(i - 1) * 2 / test_size, i = 1, test_size)]
  !ytest = gaussian_fit(xtest)
  do i = 1, test_size
    xtest(i) =  (xtest(i)-ave_x_train)/(2*sd_x_train)
    ytest(i) =  (ytest(i)-ave_y_train)/(2*sd_y_train)
  end do


  ! Instantiate a network and copy an instance to the rest of the array
  net(1) = network([input(1),&
                    dense(2,  activation=elu()),&
                    dense(10, activation=elu()),&
                    dense(64, activation=elu()),&
                    dense(96, activation=elu()),&
                    dense(64, activation=elu()),&
                    dense(10, activation=elu()),&
                    dense(1 , activation=linear())])
!  net(1) = network([input(1), dense(3), dense(1)])

  net(2:) = net(1)

  ! Print network info to stdout; this will be the same for all three networks.
  call net(1) % print_info()

!  ! SGD, no momentum
!  call sgd_optimizer(net(1), x, y, xtest, ytest, learning_rate, num_epochs)

  ! RMSProp optimizer
  call rmsprop_optimizer( &
    net(6), x, y, xtest, ytest, learning_rate, num_epochs, decay_rate &
  )




   
 end subroutine run

 subroutine Finalize()

 end subroutine finalize
!!!!!!!!!!!!!!!!!!
  subroutine rmsprop_optimizer( &
    net, x, y, xtest, ytest, learning_rate, num_epochs, decay_rate &
  )
    ! RMSprop optimizer for updating weights using root mean square
    type(network), intent(inout) :: net
    real, intent(in) :: x(:), y(:)
    real, intent(in) :: xtest(:), ytest(:)
    real, intent(in) :: learning_rate, decay_rate
    integer, intent(in) :: num_epochs
    integer :: i,j, n,lrec,irec
    real :: ypred(size(xtest))
    real :: xx(1), yy(1), ss

    print '(a)', 'RMSProp optimizer'
    print '(34("-"))'
      call net % update_dump( &
        rmsprop(learning_rate=learning_rate, decay_rate=decay_rate))

    do n = 1, num_epochs

      do i = 1, size(x)
        call net % forward([x(i)])
        call net % backward([y(i)])
      end do

      call net % update( &
        rmsprop(learning_rate=learning_rate, decay_rate=decay_rate) &
      )


    end do

      call net % update_dump( &
        rmsprop(learning_rate=learning_rate, decay_rate=decay_rate))

STOP
        DO i = 1, size(xtest)
           xx=xtest(i)
           yy= net % predict(xx)
           ypred(i)=yy(1)
           print*,xtest(i),ypred(i)*(2*sd_y_train) + ave_y_train,ytest(i)*(2*sd_y_train) + ave_y_train,ypred(i)-ytest(i)
        END DO 
        print '("Epoch: ", i4,"/",i4,", RMSE = ", f9.6)', &
             n, num_epochs, sum((ypred - ytest)**2) / size(ytest)

    INQUIRE(IOLENGTH=lrec)ypred(1) 
    open(unit=10,file='teste.bin',FORM='UNFORMATTED',&
             ACCESS='DIRECT',RECL=lrec,ACTION='WRITE',STATUS='UNKNOWN') 

    irec=0
    DO i = 1, size(xtest)
       irec=irec+1
       WRITE(10,rec=irec)ypred(i)*(2*sd_y_train) + ave_y_train
       irec=irec+1
       WRITE(10,rec=irec)ytest(i)*(2*sd_y_train) + ave_y_train
    END DO 
    print *, ''

  end subroutine rmsprop_optimizer




!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  real elemental function gaussian_fit(x) result(y)
    ! Quadratic function
    real, intent(in) :: x
    !y = (x**2 / 2 + x / 2 + 1) / 2
    y = 0.002 + 3.2e-2 * exp (- (( x - 0.0)**2)/(2*(0.007696)**2))
  end function gaussian_fit

  subroutine sgd_optimizer( &
    net, x, y, xtest, ytest, learning_rate, num_epochs, momentum, nesterov &
  )
    ! In the stochastic gradient descent (SGD) optimizer, we run the forward
    ! and backward passes and update the weights for each training sample,
    ! one at a time.
    type(network), intent(inout) :: net
    real, intent(in) :: x(:), y(:)
    real :: xx(1), yy(1)
    real, intent(in) :: xtest(:), ytest(:)
    real, intent(in) :: learning_rate
    integer, intent(in) :: num_epochs
    real, intent(in), optional :: momentum
    logical, intent(in), optional :: nesterov
    real :: ypred(size(xtest))
    real :: momentum_value
    logical :: nesterov_value
    integer :: i, n,lrec,irec

    print '(a)', 'Stochastic gradient descent'
    print '(34("-"))'

    ! Set default values for momentum and nesterov
    if (.not. present(momentum)) then
      momentum_value = 0.0
    else
      momentum_value = momentum
    end if

    if (.not. present(nesterov)) then
      nesterov_value = .false.
    else
      nesterov_value = nesterov
    end if

    do n = 1, num_epochs
      do i = 1, size(x)
        call net % forward([x(i)])
        call net % backward([y(i)])
        call net % update( &
          sgd( &
            learning_rate=learning_rate, &
            momentum=momentum_value, &
            nesterov=nesterov_value &
          ) &
        )
      end do

    end do
        DO i = 1, size(xtest)
           xx=xtest(i)
           yy= net % predict(xx)
           ypred(i)=yy(1)
           print*,xtest(i),ypred(i)*(2*sd_y_train) + ave_y_train,ytest(i)*(2*sd_y_train) + ave_y_train,ypred(i)-ytest(i)
        END DO 
        print '("Epoch: ", i4,"/",i4,", RMSE = ", f9.6)', &
             n, num_epochs, sum((ypred - ytest)**2) / size(ytest)

    print *, ''
    INQUIRE(IOLENGTH=lrec)ypred(1) 
    open(unit=10,file='teste.bin',FORM='UNFORMATTED',&
             ACCESS='DIRECT',RECL=lrec,ACTION='WRITE',STATUS='UNKNOWN') 

    irec=0
    DO i = 1, size(xtest)
       irec=irec+1
       WRITE(10,rec=irec)ypred(i)*(2*sd_y_train) + ave_y_train
       irec=irec+1
       WRITE(10,rec=irec)ytest(i)*(2*sd_y_train) + ave_y_train
    END DO 

  end subroutine sgd_optimizer



  subroutine shuffle(arr)
    ! Shuffle an array using the Fisher-Yates algorithm.
    integer, intent(inout) :: arr(:)
    real :: a
    integer :: i, temp

    do i = size(arr), 2, -1
      call random_number(a)
      a = floor(a * real(i)) + 1
      temp = arr(i)
      arr(i) = arr(int(a))
      arr(int(a)) = temp
    end do
  end subroutine shuffle

end module p2fi
