program neural_network_regression
  implicit none
  
  ! Parameters
  integer, parameter :: num_input_features = 2
  integer, parameter :: num_hidden_units = 5
  integer, parameter :: num_output_units = 1
  integer, parameter :: num_training_samples = 1000
  integer, parameter :: num_epochs = 1000
  real(kind=8), parameter :: learning_rate = 0.01
  
  ! Neural Network Weights and Biases
  real(kind=8), dimension(num_input_features, num_hidden_units) :: weights_hidden
  real(kind=8), dimension(num_hidden_units) :: biases_hidden
  real(kind=8), dimension(num_hidden_units, num_output_units) :: weights_output
  real(kind=8), dimension(num_output_units) :: biases_output
  
  ! Training Data
  real(kind=8), dimension(num_training_samples, num_input_features) :: input_data
  real(kind=8), dimension(num_training_samples, num_output_units) :: target_data
  
  ! Neural Network Outputs
  real(kind=8), dimension(num_training_samples, num_output_units) :: predictions
  
  ! Other variables
  integer :: i, j, k
  real(kind=8) :: dot_product
  real(kind=8) :: output_error
  real(kind=8) :: hidden_error  
  ! Initialize weights and biases with random values
  call random_number(weights_hidden)
  call random_number(biases_hidden)
  call random_number(weights_output)
  call random_number(biases_output)
  
  ! Generate random training data and targets (for demonstration purposes)
  do i = 1, num_training_samples
    do j = 1, num_input_features
      call random_number(input_data(i, j))
    end do
    ! Generate target values (sum of input features)
    target_data(i, 1) = sum(input_data(i, :))
  end do
  
  ! Training the neural network
  do i = 1, num_epochs
    ! Forward propagation
    do j = 1, num_training_samples
      do k = 1, num_hidden_units
        predictions(j, k) = dot_product(input_data(j, :), weights_hidden(:, k)) + biases_hidden(k)
       ! predictions(j, k) = dot_product
      end do
      do k = 1, num_output_units
        predictions(j, k) = dot_product(predictions(j, :), weights_output(:, k)) + biases_output(k)
        !predictions(j, k) = dot_product
      end do
    end do
    
    ! Backward propagation and weight updates (gradient descent)
    do j = 1, num_training_samples
      do k = 1, num_output_units
        ! Compute output layer error
!        real(kind=8) :: output_error
        output_error = predictions(j, k) - target_data(j, k)
        
        ! Update output layer weights and biases
        weights_output(:, k) = weights_output(:, k) - learning_rate * output_error * predictions(j, :)
        biases_output(k) = biases_output(k) - learning_rate * output_error
        
        ! Update hidden layer weights and biases
        do k = 1, num_hidden_units
!          real(kind=8) :: hidden_error
          hidden_error = output_error * weights_output(k, 1)
          weights_hidden(:, k) = weights_hidden(:, k) - learning_rate * hidden_error * input_data(j, :)
          biases_hidden(k) = biases_hidden(k) - learning_rate * hidden_error
        end do
      end do
    end do
  end do
  
  ! Print final weights and biases
  print *, "Final Weights (Hidden Layer):"
  do i = 1, num_input_features
    print *, weights_hidden(i, :)
  end do
  print *, "Final Biases (Hidden Layer):", biases_hidden
  
  print *, "Final Weights (Output Layer):"
  do i = 1, num_hidden_units
    print *, weights_output(i, :)
  end do
  print *, "Final Biases (Output Layer):", biases_output
  
end program neural_network_regression
